/* xmms-winamp
   implements an XMMS visualization plugin
   
   based on main.c from Julien Carme's Infinity plugin
   
   (C) 2002 Ulrich Hecht
   Licensed under the terms of the GNU General Public License v2
 */

#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/wait.h>

#include <xmms/plugin.h>
#include <xmms/configfile.h>

#include <gtk/gtk.h>

#include "winampmsg.h"
#include "plugin.h"

int finished;
short pcm_data[2][512];
short freq_data[2][256];

static GtkWidget *configure_win = NULL;
dllconfig_t config;


VisPlugin dll_vp = {
	NULL,
	NULL,
	0, /* XMMS Session ID, filled in by XMMS */
	"Winamp Meta-Plugin", /* description */
	2, /* Number of PCM channels wanted */
	2, /* Number of freq channels wanted */
	plug_init, /* init */
	plug_cleanup, /* cleanup */
	plug_about, /* about */
	plug_configure, /* configure */
	NULL, /* disable_plugin */
	NULL, /* playback_start */
	NULL, /* playback_stop */
	plug_render_pcm, /* render_pcm */
	plug_render_freq  /* render_freq */
};

VisPlugin *get_vplugin_info(void)
{
	return &dll_vp;
}

int child=0; 
int shmemfile;
dllmsg* shmem=NULL;

extern char **environ;

int prefs_loaded=0;

static void plug_init(void)
{
	char shmname[128];
	int i;

	if(!prefs_loaded)
	{
		plug_load_prefs();
		prefs_loaded=1;
	}

	sprintf(shmname,"/tmp/winamp");		/* FIXME: this should be process-specific */
	shmemfile=open(shmname,O_CREAT|O_RDWR|O_TRUNC|O_APPEND,0600);
	for(i=0;i<(sizeof(dllmsg)>>7)+1;i++) write(shmemfile,shmname,128);
	shmem=mmap(NULL,sizeof(dllmsg),PROT_READ|PROT_WRITE,MAP_SHARED,shmemfile,0);
	if(!shmem)
	{
		fprintf(stderr,"could not map shmem\n");
		dll_vp.disable_plugin(&dll_vp);
		return;
	}
	memset(shmem,0,sizeof(dllmsg));
	shmem->render=0;
	shmem->quit=0;
	shmem->ready=0x407d00d;

	child=fork();
	if(!child)
	{
		char cmdline[255];
		char* argv[3];
		setpgrp();	/* need a process group to be able to signal all potential grandchildren */
		argv[0]=WINAMPEXEPATH "/winamp";
		argv[1]=NULL;
		sprintf(cmdline,"%s/%s",config.plugindir,config.plugin);
		setenv("DLLNAME",cmdline,1);
		sprintf(cmdline,"%d",config.moduleno);
		setenv("DLLMODNUM",cmdline,1);
		execve(argv[0],argv,environ);
		exit(0);
	}
}	

static void plug_cleanup(void)
{
	int status=0;
	int i=0;
	shmem->quit=1;
	munmap(shmem,sizeof(dllmsg));
	shmem=NULL;
	close(shmemfile);

	/* wait for the plugin helper to exit */
	for(i=0;i<4;i++)
	{
		if(waitpid(child,&status,WNOHANG)==child) break;
		sleep(1);
	}

	/* make sure all grandchildren are killed as well */
	kill(-child,SIGKILL);
	waitpid(child,&status,0);
}

static void plug_render_pcm(short data[2][512])
{
	int i;
	for(i=0;i<576;i++)
	{
		shmem->pcm[0][i]=((data[0][i*512/576])>>8);
		shmem->pcm[1][i]=((data[1][i*512/576])>>8);
		//printf("%d/%d ",shmem->pcm[0][i],shmem->pcm[1][i]);
	}
	//printf("\n");
	shmem->render=1;
}

static void plug_render_freq(short data[2][256])
{
	int i;
	for(i=0;i<576;i++)
	{
		shmem->freq[0][i]=((data[0][i*256/576])>>4);
		shmem->freq[1][i]=((data[1][i*256/576])>>4);
		//printf("%d/%d ",shmem->freq[0][i],shmem->freq[1][i]);
	}
	//printf("\n");
	shmem->render=1;
}

GtkWidget* direntry=NULL;
GtkWidget* dllmenu=NULL;
GtkWidget* dllmenu_i=NULL;
GtkWidget* modmenu=NULL;
GtkWidget* modmenu_i=NULL;
		
void plug_configure(void)
{
	GtkWidget *buttonbox,*ok,*cancel,*configure,*browsedir;
	static GtkWidget *vbox,*vbox2;
      
	GtkWidget *label;
	static GtkWidget *hbox,*frame;
	
        if(configure_win)
                return;

	if(!prefs_loaded)
	{
		plug_load_prefs();
		prefs_loaded=1;
	}
	
        configure_win = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(configure_win, 558, 210);

        gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);
        gtk_window_set_title(GTK_WINDOW(configure_win), "Winamp Meta-Plugin Configuration");
        gtk_window_set_policy(GTK_WINDOW(configure_win), FALSE, FALSE, FALSE);
        gtk_window_set_position(GTK_WINDOW(configure_win), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(configure_win), "destroy", 
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &configure_win);

	vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(configure_win),vbox);

	frame = gtk_frame_new("Display:");
	gtk_box_pack_start(GTK_BOX(vbox),frame,TRUE,TRUE,4);

	vbox2 = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(frame),vbox2);
	
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,4);

	label = gtk_label_new("Plugin Directory:");
	gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,4);
	gtk_widget_show(label);

	direntry=gtk_entry_new();
	gtk_entry_set_text((GtkEntry*)direntry,config.plugindir);
	gtk_widget_show(direntry);
	gtk_box_pack_start(GTK_BOX(hbox),direntry,TRUE,TRUE,4);

	browsedir=gtk_button_new_with_label("Browse");
	gtk_widget_show(browsedir);
	gtk_box_pack_end(GTK_BOX(hbox),browsedir,FALSE,FALSE,8);

	gtk_widget_show(hbox);

	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,4);

	label = gtk_label_new("Plugin DLL:");
	gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,4);
	gtk_widget_show(label);

	recreate_dllmenu();

	gtk_box_pack_start(GTK_BOX(hbox),dllmenu,TRUE,TRUE,4);

	gtk_widget_show(hbox);

	recreate_modmenu();

	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,4);

	label = gtk_label_new("Plugin Module:");
	gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,4);
	gtk_widget_show(label);

	gtk_box_pack_start(GTK_BOX(hbox),modmenu,TRUE,TRUE,4);

	gtk_widget_show(hbox);

	gtk_signal_connect(GTK_OBJECT(direntry),"changed",
			GTK_SIGNAL_FUNC(dentry_changed),config.plugindir);
	
	gtk_widget_show(vbox2);

	gtk_widget_show(frame);
	
	buttonbox=gtk_hbutton_box_new();
	gtk_box_pack_end(GTK_BOX(vbox), buttonbox, FALSE, FALSE, 8);
	gtk_hbutton_box_set_layout_default(GTK_BUTTONBOX_END);
	gtk_widget_show(buttonbox);
	
	configure=gtk_button_new_with_label("Configure");
	GTK_WIDGET_SET_FLAGS(configure, GTK_CAN_DEFAULT);
	gtk_box_pack_end(GTK_BOX(buttonbox), configure, FALSE, FALSE, 8);
	gtk_widget_show(configure);

	ok=gtk_button_new_with_label("Ok");
	GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
	gtk_box_pack_end(GTK_BOX(buttonbox), ok, FALSE, FALSE, 8);
	gtk_widget_show(ok);

	cancel=gtk_button_new_with_label("Cancel");
	GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
	gtk_box_pack_end(GTK_BOX(buttonbox), cancel, FALSE, FALSE, 8);
	gtk_widget_show(cancel);
	
	gtk_widget_show(vbox);

	gtk_window_set_default(GTK_WINDOW(configure_win), ok);
	gtk_signal_connect(GTK_OBJECT(cancel), "clicked", GTK_SIGNAL_FUNC(cancel_clicked), configure_win);
	gtk_signal_connect(GTK_OBJECT(ok), "clicked", GTK_SIGNAL_FUNC(ok_clicked), configure_win);
	gtk_signal_connect(GTK_OBJECT(configure), "clicked", GTK_SIGNAL_FUNC(configure_clicked), configure_win);
	gtk_signal_connect(GTK_OBJECT(browsedir), "clicked", GTK_SIGNAL_FUNC(browsedir_clicked), configure_win);
	
	gtk_widget_show(configure_win);
}

static void plug_about(void)
{
	static GtkWidget *window=NULL;
	GtkWidget *vbox, *buttonbox, *close, *label;

	if(window)
		return;

	window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), NULL);
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);

	vbox=gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
	gtk_widget_show(vbox);

	label=gtk_label_new("\n\
Winamp Meta-Plugin v " VERSION " \n\n\
(C) 2002, Ulrich Hecht.\n\
e-Mail: <uli@emulinks.de> \n\
\n\
This plugin allows you to use Winamp visualization plugins with XMMS.\n\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License version 2 as\n\
published by the Free Software Foundation.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.");
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 8);
	gtk_widget_show(label);
	
	buttonbox=gtk_hbutton_box_new();
	gtk_box_pack_end(GTK_BOX(vbox), buttonbox, FALSE, FALSE,8);
	gtk_widget_show(buttonbox);
	
	close=gtk_button_new_with_label("Close");
	GTK_WIDGET_SET_FLAGS(close, GTK_CAN_DEFAULT);
	gtk_window_set_default(GTK_WINDOW(window), close);
	gtk_hbutton_box_set_layout_default(GTK_BUTTONBOX_END);
	gtk_box_pack_end(GTK_BOX(buttonbox), close, FALSE, FALSE,8);
	gtk_widget_show(close);
	
	gtk_signal_connect(GTK_OBJECT(close), "clicked", GTK_SIGNAL_FUNC(about_close_clicked), &window);
	gtk_signal_connect(GTK_OBJECT(window), "delete-event", GTK_SIGNAL_FUNC(about_closed), &window);
		
	gtk_widget_show(window);

}


static void ok_clicked(GtkWidget *w, GtkWidget *window)
{
    plug_save_prefs();
    gtk_widget_destroy(window);
    configure_win=NULL;
    modmenu=modmenu_i=NULL;
    dllmenu=dllmenu_i=NULL;
}

static void configure_clicked(GtkWidget* w, GtkWidget* window)
{
	if(shmem)
		shmem->configure=1;
}

static void dentry_changed(GtkWidget*w,char* d)
{
	strcpy(d,gtk_entry_get_text((GtkEntry*)w));
	recreate_dllmenu();
	recreate_modmenu();
}

static void dllmenu_changed(GtkWidget*w,char* v)
{
	strcpy(config.plugin,v);
	recreate_modmenu();
}

static void modmenu_changed(GtkWidget*w,char* v)
{
	config.moduleno=(int)(long)v;
}

static void about_closed(GtkWidget *w, GdkEvent *e, GtkWidget **window)
{
	about_close_clicked(w,window);
}

static void about_close_clicked(GtkWidget *w, GtkWidget **window)
{
	gtk_widget_destroy(*window);
	*window=NULL;
}

void plug_save_prefs(void)
{
    char *name;
    ConfigFile *f;
    
    name=g_strdup_printf("%s%s",g_get_home_dir(), "/.xmms/config");
    
    if(!(f=xmms_cfg_open_file(name)))
	f=xmms_cfg_new();
    
    xmms_cfg_write_string(f, "winampmeta", "version", VERSION);	
    xmms_cfg_write_string(f,"winampmeta","dir",config.plugindir);
    xmms_cfg_write_string(f,"winampmeta","dll",config.plugin);
    xmms_cfg_write_int(f,"winampmeta","module",config.moduleno);
    
    xmms_cfg_write_file(f, name);
    xmms_cfg_free(f);
    
    g_free(name);
}

void plug_load_prefs(void)
{
    char *name;
    char *someptr;
    ConfigFile *f;
    name=g_strdup_printf("%s%s",g_get_home_dir(), "/.xmms/config");

    if((f=xmms_cfg_open_file(name))) {
	if(check_cfg_version(f))
	{
	    xmms_cfg_read_string(f, "winampmeta", "dir", &someptr);
	    strcpy(config.plugindir,someptr);
	    xmms_cfg_read_string(f, "winampmeta", "dll", &someptr);
	    strcpy(config.plugin,someptr);
	    xmms_cfg_read_int(f,"winampmeta","module",&config.moduleno);
	}
	else
	{
		strcpy(config.plugindir,g_get_home_dir());
		strcpy(config.plugin,"");
		config.moduleno=0;
	}
	xmms_cfg_free(f);
    }
    g_free(name);
}

static gboolean check_cfg_version(ConfigFile *f)
{
    char *vstr;
    
    if(xmms_cfg_read_string(f, "winampmeta", "version", &vstr))
	if(!strcmp(vstr, VERSION))
	    return TRUE;
    
    return FALSE;
}

GtkWidget *file_selector;

void store_filename(GtkFileSelection *selector, gpointer user_data) {
   strcpy(config.plugindir,gtk_file_selection_get_filename (GTK_FILE_SELECTION(file_selector)));
   gtk_entry_set_text((GtkEntry*)direntry,config.plugindir);
}

void create_file_selection(void) {

   /* Create the selector */
   
   file_selector = gtk_file_selection_new("Please select a file for editing.");
   gtk_window_set_modal((GtkWindow*)file_selector,TRUE);
   gtk_file_selection_set_filename((GtkFileSelection*)file_selector,config.plugindir);
   
   gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
   			   "clicked", GTK_SIGNAL_FUNC (store_filename), NULL);
   			   
   /* Ensure that the dialog box is destroyed when the user clicks a button. */
   
   gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
   					  "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
   					  (gpointer) file_selector);

   gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->cancel_button),
   					  "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
   					  (gpointer) file_selector);
   
   /* Display that dialog */
   
   gtk_widget_show (file_selector);
}


static void browsedir_clicked(GtkWidget *w, GtkWidget *window)
{
	create_file_selection();
}

static void cancel_clicked(GtkWidget *w, GtkWidget *window)
{
	gtk_widget_destroy(window);
	configure_win=NULL;
}

char* dll_list[1000];

static char** plug_get_list(char* dir,char* file,char* cmd)
{
	FILE* listf;
	int i=0;
	char cmdline[255];
	sprintf(cmdline,"DLLNAME=\"%s/%s\" DLLCMD=%s " WINAMPEXEPATH "/winamp" ,dir,file,cmd);
	listf=popen(cmdline,"r");
	while(!feof(listf))
	{
		if(!fgets(cmdline,255,listf))
			break;
		else
		{
			//printf("%s\n",cmdline);
			if(!strncmp(cmdline,"WALIST",6))
			{
				dll_list[i]=strdup(cmdline+6);
				dll_list[i][strlen(cmdline+6)-1]=0;
				i++;
			}
		}
	}
	pclose(listf);
	dll_list[i]=NULL;
	return dll_list;
}

static void plug_free_list(char** lst)
{
	while(*lst)
		free(*(lst++));
}

static void recreate_modmenu(void)
{
	int i;
	int reset_modmenu=0;
	char** dlls;
	char** dlls_iter;
	GtkWidget* item;
	
	if(!modmenu_i)
		modmenu_i=gtk_menu_new();
	else
	{
		gtk_widget_destroy(modmenu_i);
		modmenu_i=gtk_menu_new();
		if(modmenu) {
			reset_modmenu=1;
		}
	}
	
	dlls=dlls_iter=plug_get_list(config.plugindir,config.plugin,"mods");
	i=0;
	while(*dlls_iter)
	{
		item = gtk_menu_item_new_with_label(*dlls_iter);
                gtk_object_set_data(GTK_OBJECT(item), "cmd", (gpointer)(*dlls_iter));
                gtk_widget_show(item);
                gtk_menu_append(GTK_MENU(modmenu_i), item);
                gtk_signal_connect(GTK_OBJECT(item), "activate", modmenu_changed, (gpointer)i);
                dlls_iter++,i++;
        }
        plug_free_list(dlls);

	if(config.moduleno>=i)
		config.moduleno=0;
        
        if(!modmenu) {
        	modmenu = gtk_option_menu_new();
	        gtk_widget_set_usize(modmenu, 120, -1);
	        gtk_option_menu_remove_menu(GTK_OPTION_MENU(modmenu));
	        gtk_option_menu_set_menu(GTK_OPTION_MENU(modmenu), modmenu_i);
	        gtk_object_set_data(GTK_OBJECT(modmenu), "modmenu_i", (gpointer)modmenu_i);
		gtk_option_menu_set_history(GTK_OPTION_MENU(modmenu),config.moduleno);
	        gtk_widget_show(modmenu);
	}
	if(reset_modmenu)
	{
		gtk_option_menu_set_menu(GTK_OPTION_MENU(modmenu), modmenu_i);
		gtk_option_menu_set_history(GTK_OPTION_MENU(modmenu),config.moduleno);
	}
}

static void recreate_dllmenu(void)
{
	int i,defval=-1;
	int reset_dllmenu=0;
	char** dlls;
	char** dlls_iter;
	char* dlldesc;
	GtkWidget* item;
	char buffer[255];
	
	if(!dllmenu_i)
		dllmenu_i=gtk_menu_new();
	else
	{
		gtk_widget_destroy(dllmenu_i);
		dllmenu_i=gtk_menu_new();
		if(dllmenu) {
			reset_dllmenu=1;
		}
	}
	
	dlls=dlls_iter=plug_get_list(config.plugindir,"","plugs");
	i=0;
	while(*dlls_iter)
	{
		dlldesc=strstr(*dlls_iter,",,")+2;
		*(dlldesc-2)=0;
		sprintf(buffer,"%s [%s]",dlldesc,*dlls_iter);
		item = gtk_menu_item_new_with_label(buffer);
                gtk_object_set_data(GTK_OBJECT(item), "cmd", (gpointer)(buffer));
                gtk_widget_show(item);
                gtk_menu_append(GTK_MENU(dllmenu_i), item);
                gtk_signal_connect(GTK_OBJECT(item), "activate", dllmenu_changed, strdup(*dlls_iter));
                if(!strcmp(*dlls_iter,config.plugin))
                	defval=i;
                dlls_iter++,i++;
        }
        if(defval==-1)
        {
        	if(*dlls)
        		strcpy(config.plugin,*dlls);
        	else
        		config.plugin[0]=0;
        	defval=0;
        }
        plug_free_list(dlls);
        
        if(!dllmenu)
        {
	        dllmenu = gtk_option_menu_new();
        	gtk_widget_set_usize(dllmenu, 120, -1);
	        gtk_option_menu_remove_menu(GTK_OPTION_MENU(dllmenu));
        	gtk_option_menu_set_menu(GTK_OPTION_MENU(dllmenu), dllmenu_i);
	        gtk_object_set_data(GTK_OBJECT(dllmenu), "dllmenu_i", (gpointer)dllmenu_i);
		gtk_option_menu_set_history(GTK_OPTION_MENU(dllmenu), defval);
	        gtk_widget_show(dllmenu);
	}
	if(reset_dllmenu)
	{
		gtk_option_menu_set_menu(GTK_OPTION_MENU(dllmenu), dllmenu_i);
		gtk_option_menu_set_history(GTK_OPTION_MENU(dllmenu),defval);
	}
}
