/* xmms-winamp
   plugin helper binary, interface to Winamp plugin
   
   (C) 2002 Ulrich Hecht
   Licensed under the terms of the GNU General Public License v2
 */

/* #include <unistd.h> it seems that windows.h and unistd.h don't go too well together */
extern void usleep(long);
extern int close(int);
extern int chdir(const char*);

#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <libgen.h>

#include <windows.h>
#include "vis.h"
#include "winampmsg.h"

extern void list_plugins(void);
extern void list_modules(void);

int WINAPI WinMain(HINSTANCE eins, HINSTANCE zwei, LPSTR arg, int whatever)
{
	HINSTANCE dll;
	winampVisHeader* header;
	winampVisModule* module;
	
	int shmfile;
	int result;
	int modnum;
	
	dllmsg* shmem;

	char* cmd;
	
	winampVisGetHeaderType headerfunc=0;
	
	cmd=getenv("DLLCMD");
	if(cmd)
	{
		if(!strcmp(cmd,"plugs"))
		{
			list_plugins();
			return 0;
		}
		if(!strcmp(cmd,"mods"))
		{
			list_modules();
			return 0;
		}
	}

	printf("opening %s\n",getenv("DLLNAME"));
	dll=LoadLibrary(getenv("DLLNAME"));
	
	if(!dll)
	{
		fprintf(stderr,"could not load dll %s\n",getenv("DLLNAME"));
		exit(1);
	}
	
	headerfunc=(winampVisGetHeaderType)GetProcAddress(dll,"winampVisGetHeader");
	
	header=headerfunc();
	
	printf("version: %d, description: %s\n",header->version,header->description);
	
	modnum=atoi(getenv("DLLMODNUM"));
	printf("module number %d requested\n",modnum);
	module=header->getModule(modnum);
	
	module->hwndParent=0;	/* we don't have a main Windoze window */
	module->hDllInstance=dll;
	module->sRate=44100;		/* FIXME needs to come from xmms, but it doesn't seem to supply anything like that */
	module->nCh=2;			/* same here */
	
	
	if((result=module->Init(module)))
	{
		printf("Error no. %d in Init()\n",result);
		exit(43);
	}
	
	shmfile=open("/tmp/winamp",O_RDWR);
	shmem=mmap(NULL,sizeof(dllmsg),PROT_READ|PROT_WRITE,MAP_SHARED,shmfile,0);
	if(!shmem)
	{
		fprintf(stderr,"could not mmap shmem in dll\n");
		exit(45);
	}

	shmem->pcmwanted=module->spectrumNch;
	shmem->freqwanted=module->waveformNch;

	while(shmem->ready!=0x407d00d)
	{
		usleep(100000);
	}

	for(;;)
	{
		MSG msg;
		if(shmem->render)
		{
			memcpy(module->spectrumData,shmem->freq,2*576);
			memcpy(module->waveformData,shmem->pcm,2*576);
			//printf("render frame\n");
			shmem->render=0;
			module->Render(module);
			continue;
		}
		if(shmem->configure)
		{
			module->Config(module);
			shmem->configure=0;
			continue;
		}
		if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) shmem->quit = 1;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		if(shmem->quit)
			break;
		
		usleep(10000);
	}
	
	module->Quit(module);
	munmap(shmem,sizeof(dllmsg));
	close(shmfile);
	
	FreeLibrary(dll);
	
	return 0;
}
